@@ -23,13 +23,18 @@
 #include "workspace.h"
 #include "window.h"
 #include "window-action-menu.h"
+#include "xutils.h"
 
-#define N_SCREEN_CONNECTIONS 9
+#define N_SCREEN_CONNECTIONS 10
 
 struct _WnckPagerPrivate
 {
   WnckScreen *screen;
+  
   int n_rows; /* really columns for vertical orientation */
+  WnckPagerDisplayMode display_mode;
+  gboolean show_all_workspaces;
+  
   GtkOrientation orientation;
   int workspace_size;
   guint screen_connections[N_SCREEN_CONNECTIONS];
@@ -44,10 +49,13 @@ struct _WnckPagerPrivate
   int action_click_x;
   int action_click_y;
   guint dragging : 1;
+
+  GdkPixbuf *bg_cache;
 };
 
 enum
 {
+  dummy, /* remove this when you add more signals */
   LAST_SIGNAL
 };
 
@@ -87,6 +95,10 @@ static void wnck_pager_disconnect_screen (WnckPager  *pager);
 
 static void wnck_pager_clear_drag (WnckPager *pager);
 
+static GdkPixbuf* wnck_pager_get_background (WnckPager *pager,
+                                             int        width,
+                                             int        height);
+
 static gpointer parent_class;
 static guint signals[LAST_SIGNAL] = { 0 };
 
@@ -126,8 +138,11 @@ wnck_pager_init (WnckPager *pager)
   pager->priv = g_new0 (WnckPagerPrivate, 1);
 
   pager->priv->n_rows = 1;
+  pager->priv->display_mode = WNCK_PAGER_DISPLAY_CONTENT;
+  pager->priv->show_all_workspaces = TRUE;
   pager->priv->orientation = GTK_ORIENTATION_HORIZONTAL;
   pager->priv->workspace_size = 48;
+  pager->priv->bg_cache = NULL;
 }
 
 static void
@@ -159,6 +174,12 @@ wnck_pager_finalize (GObject *object)
   pager = WNCK_PAGER (object);
 
   wnck_pager_disconnect_screen (pager);
+
+  if (pager->priv->bg_cache)
+    {
+      g_object_unref (G_OBJECT (pager->priv->bg_cache));
+      pager->priv->bg_cache = NULL;
+    }
   
   g_free (pager->priv);
   
@@ -220,6 +241,7 @@ wnck_pager_size_request  (GtkWidget      *widget,
   int other_dimension_size;
   int u_width, u_height;
   int size;
+  int n_rows;
   
   pager = WNCK_PAGER (widget);
   
@@ -232,27 +254,85 @@ wnck_pager_size_request  (GtkWidget      *widget,
   
   if (pager->priv->orientation == GTK_ORIENTATION_VERTICAL)
     {
-      size = pager->priv->workspace_size;
-      if (u_width != -1)
-	size = (u_width - (pager->priv->n_rows - 1))  / pager->priv->n_rows;
-      
       screen_aspect = (double) gdk_screen_height () / (double) gdk_screen_width ();
+
+      /* TODO: Handle WNCK_PAGER_DISPLAY_NAME for this case */
+
+      if (pager->priv->show_all_workspaces)
+	{
+	  size = pager->priv->workspace_size;
+	  if (u_width != -1)
+	    size = (u_width - (pager->priv->n_rows - 1))  / pager->priv->n_rows;
+ 
+	  n_rows = pager->priv->n_rows;
+	}
+      else
+	{
+	  size = pager->priv->workspace_size;
+	  if (u_width != -1)
+	    size = u_width;
+
+	  n_rows = 1;
+	  spaces_per_row = 1;
+	}
+      
       other_dimension_size = screen_aspect * size;
       
-      requisition->width = size * pager->priv->n_rows + (pager->priv->n_rows - 1);
+      requisition->width = size * n_rows + (n_rows - 1);
       requisition->height = other_dimension_size * spaces_per_row  + (spaces_per_row - 1);
     }
   else
     {
-      size = pager->priv->workspace_size;
-      if (u_height != -1)
-	size = (u_height - (pager->priv->n_rows - 1))/ pager->priv->n_rows;
-      
       screen_aspect = (double) gdk_screen_width () / (double) gdk_screen_height ();
-      other_dimension_size = screen_aspect * size;
+      
+      if (pager->priv->show_all_workspaces)
+	{
+	  size = pager->priv->workspace_size;
+	  if (u_height != -1)
+	    size = (u_height - (pager->priv->n_rows - 1))/ pager->priv->n_rows;
+ 
+	  n_rows = pager->priv->n_rows;
+	}
+      else
+	{
+	  size = pager->priv->workspace_size;
+	  if (u_height != -1)
+	    size = u_height;
 
+	  n_rows = 1;
+	  spaces_per_row = 1;
+	}
+
+      if (pager->priv->display_mode == WNCK_PAGER_DISPLAY_CONTENT)
+	{
+	  other_dimension_size = screen_aspect * size;
+	}
+      else
+	{
+	  int n_spaces, i, w;
+	  PangoLayout *layout;
+
+	  n_spaces = wnck_screen_get_workspace_count (pager->priv->screen);
+	  layout = gtk_widget_create_pango_layout  (widget, NULL);
+	  
+	  other_dimension_size = 1;
+	  
+	  for (i = 0; i < n_spaces; i++)
+	    {
+	      pango_layout_set_text (layout,
+				     wnck_workspace_get_name (wnck_workspace_get (i)),
+				     -1);
+	      pango_layout_get_pixel_size (layout, &w, NULL);
+	      other_dimension_size = MAX (other_dimension_size, w);
+	    }
+	  
+	  g_object_unref (layout);
+	  
+	  other_dimension_size += 2;
+	}
+      
       requisition->width = other_dimension_size * spaces_per_row + (spaces_per_row - 1);
-      requisition->height = size * pager->priv->n_rows + (pager->priv->n_rows - 1);
+      requisition->height = size * n_rows + (n_rows - 1);
     }
 }
 
@@ -278,6 +358,28 @@ get_workspace_rect (WnckPager    *pager,
   int col, row;
   
   widget = GTK_WIDGET (pager);
+
+  if (!pager->priv->show_all_workspaces)
+    {
+      WnckWorkspace *active_space;
+  
+      active_space = wnck_screen_get_active_workspace (pager->priv->screen);
+
+      rect->x = 0;
+      rect->y = 0;
+      if (active_space && space == wnck_workspace_get_number (active_space))
+	{
+	  rect->width = widget->allocation.width;
+	  rect->height = widget->allocation.height;
+	}
+      else
+	{
+	  rect->width = 0;
+	  rect->height = 0;
+	}
+
+      return;
+    }
   
   n_spaces = wnck_screen_get_workspace_count (pager->priv->screen);
 
@@ -498,6 +600,103 @@ workspace_at_point (WnckPager *pager,
   return -1;
 }
 
+
+static void
+wnck_pager_draw_workspace (WnckPager    *pager,
+			   int           workspace,
+			   GdkRectangle *rect,
+                           GdkPixbuf    *bg_pixbuf)
+{
+  GList *windows;
+  GList *tmp;
+  gboolean is_current;
+  WnckWorkspace *active_space;
+  
+  active_space = wnck_screen_get_active_workspace (pager->priv->screen);
+
+  is_current = active_space &&
+    workspace == wnck_workspace_get_number (active_space);
+
+  /* FIXME in names mode, should probably draw things like a button.
+   */
+  
+  if (is_current)
+    gdk_draw_rectangle (GTK_WIDGET (pager)->window,
+                        GTK_WIDGET (pager)->style->dark_gc[GTK_STATE_SELECTED],
+                        TRUE,
+                        rect->x, rect->y, rect->width, rect->height);
+  else if (bg_pixbuf)
+    {
+      gdk_pixbuf_render_to_drawable (bg_pixbuf,
+                                     GTK_WIDGET (pager)->window,
+                                     GTK_WIDGET (pager)->style->dark_gc[GTK_STATE_SELECTED],
+                                     0, 0,
+                                     rect->x, rect->y,
+                                     -1, -1,
+                                     GDK_RGB_DITHER_MAX,
+                                     0, 0);
+    }
+  else
+    gdk_draw_rectangle (GTK_WIDGET (pager)->window,
+                        GTK_WIDGET (pager)->style->dark_gc[GTK_STATE_NORMAL],
+                        TRUE,
+                        rect->x, rect->y, rect->width, rect->height);
+
+  if (pager->priv->display_mode == WNCK_PAGER_DISPLAY_CONTENT)
+    {      
+      windows = get_windows_for_workspace_in_bottom_to_top (pager->priv->screen,
+							    wnck_workspace_get (workspace));
+      
+      tmp = windows;
+      while (tmp != NULL)
+	{
+	  WnckWindow *win = tmp->data;
+	  GdkRectangle winrect;
+	  
+	  if (pager->priv->dragging &&
+	      win == pager->priv->drag_window)
+	    {
+	      tmp = tmp->next;
+	      continue;
+	    }
+	  
+	  get_window_rect (win, rect, &winrect);
+	  
+	  draw_window (GTK_WIDGET (pager)->window,
+		       GTK_WIDGET (pager),
+		       win,
+		       &winrect);
+	  
+	  tmp = tmp->next;
+	}
+      
+      g_list_free (windows);
+    }
+  else
+    {
+      /* Workspace name mode */
+      PangoLayout *layout;
+      int w, h;
+
+      
+      
+      layout = gtk_widget_create_pango_layout  (GTK_WIDGET (pager),
+						wnck_workspace_get_name (wnck_workspace_get (workspace)));
+      
+      pango_layout_get_pixel_size (layout, &w, &h);
+      
+      gdk_draw_layout  (GTK_WIDGET (pager)->window,
+			is_current ?
+			GTK_WIDGET (pager)->style->fg_gc[GTK_STATE_SELECTED] :
+			GTK_WIDGET (pager)->style->fg_gc[GTK_STATE_NORMAL],
+			rect->x + (rect->width - w) / 2,
+			rect->y + (rect->height - h) / 2,
+			layout);
+
+      g_object_unref (layout);
+    }
+}
+
 static gboolean
 wnck_pager_expose_event  (GtkWidget      *widget,
                           GdkEventExpose *event)
@@ -506,65 +705,42 @@ wnck_pager_expose_event  (GtkWidget      *widget,
   int i;
   int n_spaces;
   WnckWorkspace *active_space;
+  GdkPixbuf *bg_pixbuf;
+  gboolean first;
   
   pager = WNCK_PAGER (widget);
 
   n_spaces = wnck_screen_get_workspace_count (pager->priv->screen);
   active_space = wnck_screen_get_active_workspace (pager->priv->screen);
+  bg_pixbuf = NULL;
+  first = TRUE;
   
   i = 0;
   while (i < n_spaces)
     {
       GdkRectangle rect;
-      GList *windows;
-      GList *tmp;
-      gboolean is_current;
-      
-      get_workspace_rect (pager, i, &rect);
-      
-      is_current = active_space &&
-        i == wnck_workspace_get_number (active_space);
 
-      if (is_current)
-        gdk_draw_rectangle (widget->window,                            
-                            widget->style->dark_gc[GTK_STATE_SELECTED],
-                            TRUE,
-                            rect.x, rect.y, rect.width, rect.height);
-      else
-	gdk_draw_rectangle (widget->window,
-                            widget->style->dark_gc[GTK_STATE_NORMAL],
-			    TRUE,
-			    rect.x, rect.y, rect.width, rect.height);
+      if (pager->priv->show_all_workspaces ||
+	  (active_space && i == wnck_workspace_get_number (active_space)))
+	{
+	  get_workspace_rect (pager, i, &rect);
 
-      
-      windows = get_windows_for_workspace_in_bottom_to_top (pager->priv->screen,
-                                                            wnck_workspace_get (i));
-      
-      tmp = windows;
-      while (tmp != NULL)
-        {
-          WnckWindow *win = tmp->data;
-          GdkRectangle winrect;
-
-          if (pager->priv->dragging &&
-              win == pager->priv->drag_window)
+          /* We only want to do this once, even if w/h change,
+           * for efficiency. width/height will only change by
+           * one pixel at most.
+           */
+          if (first &&
+              pager->priv->display_mode == WNCK_PAGER_DISPLAY_CONTENT)
             {
-              tmp = tmp->next;
-              continue;
+              bg_pixbuf = wnck_pager_get_background (pager,
+                                                     rect.width,
+                                                     rect.height);
+              first = FALSE;
             }
           
-          get_window_rect (win, &rect, &winrect);
-
-          draw_window (widget->window,
-                       widget,
-                       win,
-                       &winrect);
-          
-          tmp = tmp->next;
-        }
-
-      g_list_free (windows);
-
+	  wnck_pager_draw_workspace (pager, i, &rect, bg_pixbuf);
+	}
+ 
       ++i;
     }
 
@@ -736,9 +912,12 @@ wnck_pager_button_release (GtkWidget        *widget,
 	  space = wnck_workspace_get (i);
 
 	  if (space)
-	    wnck_window_move_to_workspace (pager->priv->drag_window,
-					   space);
-	  
+            {
+              wnck_window_move_to_workspace (pager->priv->drag_window,
+                                             space);
+              if (space == wnck_screen_get_active_workspace (pager->priv->screen))
+                wnck_window_activate (pager->priv->drag_window);
+            }	  
 	}
       
       wnck_pager_clear_drag (pager);
@@ -819,6 +998,35 @@ wnck_pager_set_n_rows (WnckPager *pager,
   gtk_widget_queue_resize (GTK_WIDGET (pager));
 }
 
+void
+wnck_pager_set_display_mode (WnckPager            *pager,
+			     WnckPagerDisplayMode  mode)
+{
+  g_return_if_fail (WNCK_IS_PAGER (pager));
+
+  if (pager->priv->display_mode == mode)
+    return;
+
+  pager->priv->display_mode = mode;
+  gtk_widget_queue_resize (GTK_WIDGET (pager));
+}
+
+void
+wnck_pager_set_show_all (WnckPager *pager,
+			 gboolean   show_all_workspaces)
+{
+  g_return_if_fail (WNCK_IS_PAGER (pager));
+
+  show_all_workspaces = (show_all_workspaces != 0);
+
+  if (pager->priv->show_all_workspaces == show_all_workspaces)
+    return;
+
+  pager->priv->show_all_workspaces = show_all_workspaces;
+  gtk_widget_queue_resize (GTK_WIDGET (pager));
+}
+
+
 static void
 active_window_changed_callback    (WnckScreen      *screen,
                                    gpointer         data)
@@ -944,6 +1152,21 @@ window_geometry_changed_callback  (WnckWindow      *window,
   gtk_widget_queue_draw (GTK_WIDGET (pager));
 }
 
+static void
+background_changed_callback (WnckWindow *window,
+                             gpointer    data)
+{
+  WnckPager *pager = WNCK_PAGER (data);
+
+  if (pager->priv->bg_cache)
+    {
+      g_object_unref (G_OBJECT (pager->priv->bg_cache));
+      pager->priv->bg_cache = NULL;
+    }
+  
+  gtk_widget_queue_draw (GTK_WIDGET (pager));
+}
+
 static void
 wnck_pager_connect_screen (WnckPager  *pager,
                            WnckScreen *screen)
@@ -1011,6 +1234,11 @@ wnck_pager_connect_screen (WnckPager  *pager,
                            pager);
   ++i;
 
+  c[i] = g_signal_connect (G_OBJECT (screen), "background_changed",
+                           G_CALLBACK (background_changed_callback),
+                           pager);
+  ++i;
+  
   g_assert (i == N_SCREEN_CONNECTIONS);
 }
 
@@ -1071,3 +1299,61 @@ wnck_pager_clear_drag (WnckPager *pager)
   pager->priv->drag_window_x = -1;
   pager->priv->drag_window_y = -1;
 }
+
+static GdkPixbuf*
+wnck_pager_get_background (WnckPager *pager,
+                           int        width,
+                           int        height)
+{
+  Pixmap p;
+  GdkPixbuf *pix = NULL;
+  
+  /* We have to be careful not to keep alternating between
+   * width/height values, otherwise this would get really slow.
+   */
+  if (pager->priv->bg_cache &&
+      gdk_pixbuf_get_width (pager->priv->bg_cache) == width &&
+      gdk_pixbuf_get_height (pager->priv->bg_cache) == height)
+    return pager->priv->bg_cache;
+
+  if (pager->priv->bg_cache)
+    {
+      g_object_unref (G_OBJECT (pager->priv->bg_cache));
+      pager->priv->bg_cache = NULL;
+    }
+
+  if (pager->priv->screen == NULL)
+    return NULL;
+
+  /* FIXME this just globally disables the thumbnailing feature */
+  return NULL;
+  
+#define MIN_BG_SIZE 10
+  
+  if (width < MIN_BG_SIZE || height < MIN_BG_SIZE)
+    return NULL;
+  
+  p = wnck_screen_get_background_pixmap (pager->priv->screen);
+
+  if (p != None)
+    {
+      _wnck_error_trap_push ();
+      pix = _wnck_gdk_pixbuf_get_from_pixmap (NULL,
+                                              p,
+                                              0, 0, 0, 0,
+                                              -1, -1);
+      _wnck_error_trap_pop ();
+    }
+
+  if (pix)
+    {
+      pager->priv->bg_cache = gdk_pixbuf_scale_simple (pix,
+                                                       width,
+                                                       height,
+                                                       GDK_INTERP_BILINEAR);
+
+      g_object_unref (G_OBJECT (pix));
+    }
+
+  return pager->priv->bg_cache;
+}
