@@ -22,6 +22,7 @@
 #include <stdio.h>
 #include "tasklist.h"
 #include "window.h"
+#include "window-action-menu.h"
 #include "workspace.h"
 #include "application.h"
 #include "xutils.h"
@@ -53,7 +54,7 @@ typedef struct _WnckTaskClass   WnckTaskClass;
 #define MINI_ICON_SIZE 16
 #define DEFAULT_GROUPING_LIMIT 80
 
-#define DEFAULT_WIDTH 300
+#define DEFAULT_WIDTH 1
 #define DEFAULT_HEIGHT 48
 
 struct _WnckTask
@@ -81,7 +82,10 @@ struct _WnckTask
   gulong name_changed_tag;
   gulong app_name_changed_tag;
 
+  /* task menu */
   GtkWidget *menu;
+  /* ops menu */
+  GtkWidget *action_menu;
 };
 
 struct _WnckTaskClass
@@ -116,6 +120,11 @@ struct _WnckTasklistPrivate
   gint grouping_limit;
 
   guint activate_timeout_id;
+
+  int *size_hints;
+  int size_hints_len;
+
+  gint minimum_width;
 };
 
 
@@ -132,6 +141,7 @@ static char      *wnck_task_get_text (WnckTask *task);
 static GdkPixbuf *wnck_task_get_icon (WnckTask *task);
 static gint       wnck_task_compare  (gconstpointer  a,
 				      gconstpointer  b);
+static void       wnck_task_update_visible_state (WnckTask *task);
 
 
 static void wnck_tasklist_init        (WnckTasklist      *tasklist);
@@ -286,6 +296,12 @@ wnck_task_finalize (GObject *object)
       task->menu = NULL;
     }
 
+  if (task->action_menu)
+    {
+      gtk_widget_destroy (task->action_menu);
+      task->action_menu = NULL;
+    }
+
   if (task->window)
     {
       g_object_unref (task->window);
@@ -346,6 +362,8 @@ wnck_tasklist_init (WnckTasklist *tasklist)
   
   tasklist->priv->grouping = WNCK_TASKLIST_AUTO_GROUP;
   tasklist->priv->grouping_limit = DEFAULT_GROUPING_LIMIT;
+
+  tasklist->priv->minimum_width = DEFAULT_WIDTH;
 }
 
 static void
@@ -395,6 +413,10 @@ wnck_tasklist_finalize (GObject *object)
       tasklist->priv->tooltips = NULL;
     }
 
+  g_free (tasklist->priv->size_hints);
+  tasklist->priv->size_hints = NULL;
+  tasklist->priv->size_hints_len = 0;
+
   g_free (tasklist->priv);
   tasklist->priv = NULL;  
   
@@ -452,68 +474,26 @@ wnck_tasklist_set_grouping_limit (WnckTasklist *tasklist,
   gtk_widget_queue_resize (GTK_WIDGET (tasklist));
 }
 
-static void
-wnck_tasklist_size_request  (GtkWidget      *widget,
-                             GtkRequisition *requisition)
+/* set the minimum width */
+void 
+wnck_tasklist_set_minimum_width (WnckTasklist *tasklist, gint size)
 {
-  WnckTasklist *tasklist;
-  GtkRequisition child_req;
-  int max_height = 1;
-  int max_width = 1;
-  int u_width, u_height;
-  GList *l;
-  
-  tasklist = WNCK_TASKLIST (widget);
+  g_return_if_fail (WNCK_IS_TASKLIST (tasklist));
 
-  /* Calculate max needed height and width of the buttons */
-  l = tasklist->priv->windows;
-  while (l != NULL)
-    {
-      WnckTask *task = WNCK_TASK (l->data);
+  if (tasklist->priv->minimum_width == size)
+    return;
 
-      gtk_widget_size_request (task->button, &child_req);
-      
-      max_height = MAX (child_req.height,
-			max_height);
-      max_width = MAX (child_req.width,
-		       max_width);
-      
-      l = l->next;
-    }
-
-  l = tasklist->priv->applications;
-  while (l != NULL)
-    {
-      WnckTask *task = WNCK_TASK (l->data);
-      
-      gtk_widget_size_request (task->button, &child_req);
-      
-      max_height = MAX (child_req.height,
-			max_height);
-      max_width = MAX (child_req.width,
-		       max_width);
-      
-      l = l->next;
-    }
-
-  tasklist->priv->max_button_width = max_width;
-  tasklist->priv->max_button_height = max_height;
-
-
-  gtk_widget_get_size_request (widget, &u_width, &u_height);
-
-  requisition->width = DEFAULT_WIDTH;
-  requisition->height = DEFAULT_HEIGHT;
-  
-  if (u_height != -1)
-    {
-      requisition->height = u_height;
-    }
-  else if (u_width != -1)
-    {
-      requisition->width = u_width;
-      requisition->height = 4 * max_height;
-    }
+  tasklist->priv->minimum_width = size;
+  gtk_widget_queue_resize (GTK_WIDGET (tasklist));
+}
+ 
+/* get the minimum width */
+gint
+wnck_tasklist_get_minimum_width (WnckTasklist *tasklist)
+{
+  g_return_val_if_fail (WNCK_IS_TASKLIST (tasklist), 0);
+	
+  return tasklist->priv->minimum_width;
 }
 
 /* returns the maximal possible button width (i.e. if you
@@ -629,7 +609,185 @@ wnck_task_get_highest_scored (GList     *ungrouped_apps,
   return g_list_remove (ungrouped_apps, best_task);
 }
 
+static void
+wnck_tasklist_size_request  (GtkWidget      *widget,
+                             GtkRequisition *requisition)
+{
+  WnckTasklist *tasklist;
+  GtkRequisition child_req;
+  GtkAllocation  fake_allocation;
+  int max_height = 1;
+  int max_width = 1;
+  int u_width, u_height;
+  GList *l;
+  GArray *array;
+  GList *ungrouped_apps;
+  int n_windows;
+  int n_rows;
+  int n_cols, last_n_cols;
+  int n_grouped_buttons;
+  gboolean score_set;
+  int val;
+  WnckTask *app_task;
+  int lowest_range;
+  int grouping_limit;
+  
+  tasklist = WNCK_TASKLIST (widget);
 
+  /* Calculate max needed height and width of the buttons */
+  l = tasklist->priv->windows;
+  while (l != NULL)
+    {
+      WnckTask *task = WNCK_TASK (l->data);
+
+      gtk_widget_size_request (task->button, &child_req);
+      
+      max_height = MAX (child_req.height,
+			max_height);
+      max_width = MAX (child_req.width,
+		       max_width);
+      
+      l = l->next;
+    }
+
+  l = tasklist->priv->applications;
+  while (l != NULL)
+    {
+      WnckTask *task = WNCK_TASK (l->data);
+      
+      gtk_widget_size_request (task->button, &child_req);
+      
+      max_height = MAX (child_req.height,
+			max_height);
+      max_width = MAX (child_req.width,
+		       max_width);
+      
+      l = l->next;
+    }
+
+  tasklist->priv->max_button_width = max_width;
+  tasklist->priv->max_button_height = max_height;
+
+
+  gtk_widget_get_size_request (widget, &u_width, &u_height);
+
+  requisition->width = tasklist->priv->minimum_width;
+  requisition->height = DEFAULT_HEIGHT;
+  
+  if (u_height != -1)
+    {
+      requisition->height = u_height;
+    }
+  else if (u_width != -1)
+    {
+      requisition->width = u_width;
+      requisition->height = 4 * max_height;
+    }
+  
+  fake_allocation.width = requisition->width;
+  fake_allocation.height = requisition->height;
+
+  array = g_array_new (FALSE, FALSE, sizeof (int));
+
+  /* Calculate size_hints list */
+  
+  n_windows = g_list_length (tasklist->priv->windows);
+  n_grouped_buttons = 0;
+  ungrouped_apps = g_list_copy (tasklist->priv->applications);
+  score_set = FALSE;
+
+  grouping_limit = MIN (tasklist->priv->grouping_limit,
+			tasklist->priv->max_button_width);
+  
+  /* Try ungrouped mode */
+  wnck_tasklist_layout (&fake_allocation,
+			tasklist->priv->max_button_width,
+			tasklist->priv->max_button_height,
+			n_windows,
+			&n_cols, &n_rows);
+
+  last_n_cols = G_MAXINT;
+  lowest_range = G_MAXINT;
+  if (tasklist->priv->grouping != WNCK_TASKLIST_ALWAYS_GROUP)
+    {
+      val = n_cols * tasklist->priv->max_button_width;
+      g_array_insert_val (array, array->len, val);
+      val = n_cols * grouping_limit;
+      g_array_insert_val (array, array->len, val);
+
+      last_n_cols = n_cols;
+      lowest_range = val;
+    }
+
+  while (ungrouped_apps != NULL &&
+	 tasklist->priv->grouping != WNCK_TASKLIST_NEVER_GROUP)
+    {
+      if (!score_set)
+	{
+	  wnck_tasklist_score_groups (tasklist, ungrouped_apps);
+	  score_set = TRUE;
+	}
+
+      ungrouped_apps = wnck_task_get_highest_scored (ungrouped_apps, &app_task);
+
+      n_grouped_buttons += g_list_length (app_task->windows) - 1;
+
+      wnck_tasklist_layout (&fake_allocation,
+			    tasklist->priv->max_button_width,
+			    tasklist->priv->max_button_height,
+			    n_windows - n_grouped_buttons,
+			    &n_cols, &n_rows);
+      if (n_cols != last_n_cols &&
+	  (tasklist->priv->grouping == WNCK_TASKLIST_AUTO_GROUP ||
+	   ungrouped_apps == NULL))
+	{
+	  val = n_cols * tasklist->priv->max_button_width;
+	  if (val >= lowest_range)
+	    { /* Overlaps old range */
+	      lowest_range = n_cols * grouping_limit;
+	      if (array->len > 0)
+		g_array_index(array, int, array->len-1) = lowest_range;
+	      else
+		g_array_insert_val (array, 0, lowest_range);
+	    }
+	  else
+	    {
+	      /* Full new range */
+	      g_array_insert_val (array, array->len, val);
+	      val = n_cols * grouping_limit;
+	      g_array_insert_val (array, array->len, val);
+	      lowest_range = val;
+	    }
+
+	  last_n_cols = n_cols;
+	}
+    }
+
+  /* Always let you go down to a zero size: */
+  if (array->len > 0)
+    g_array_index(array, int, array->len-1) = 0;
+  else
+    {
+      val = 0;
+      g_array_insert_val (array, 0, val);
+    }
+  
+  if (tasklist->priv->size_hints)
+    g_free (tasklist->priv->size_hints);
+
+  tasklist->priv->size_hints_len = array->len;
+    
+  tasklist->priv->size_hints = (int *)g_array_free (array, FALSE);
+}
+
+const int *
+wnck_tasklist_get_size_hint_list (WnckTasklist  *tasklist,
+				  int           *n_elements)
+{
+  *n_elements = tasklist->priv->size_hints_len;
+  return tasklist->priv->size_hints;
+}
+  
 static void
 wnck_tasklist_size_allocate (GtkWidget      *widget,
                              GtkAllocation  *allocation)
@@ -649,6 +807,7 @@ wnck_tasklist_size_allocate (GtkWidget      *widget,
   GList *ungrouped_apps;
   WnckTask *win_task;
   GList *visible_tasks = NULL;
+  int grouping_limit;
   
   tasklist = WNCK_TASKLIST (widget);
 
@@ -657,6 +816,9 @@ wnck_tasklist_size_allocate (GtkWidget      *widget,
   ungrouped_apps = g_list_copy (tasklist->priv->applications);
   score_set = FALSE;
 
+  grouping_limit = MIN (tasklist->priv->grouping_limit,
+			tasklist->priv->max_button_width);
+
   /* Try ungrouped mode */
   button_width = wnck_tasklist_layout (allocation,
 				       tasklist->priv->max_button_width,
@@ -666,7 +828,7 @@ wnck_tasklist_size_allocate (GtkWidget      *widget,
   while (ungrouped_apps != NULL &&
 	 ((tasklist->priv->grouping == WNCK_TASKLIST_ALWAYS_GROUP) ||
 	  ((tasklist->priv->grouping == WNCK_TASKLIST_AUTO_GROUP) &&
-	   (button_width < tasklist->priv->grouping_limit))))
+	   (button_width < grouping_limit))))
     {
       if (!score_set)
 	{
@@ -723,6 +885,7 @@ wnck_tasklist_size_allocate (GtkWidget      *widget,
   i = 0;
   total_width = tasklist->priv->max_button_width * n_cols;
   total_width = MIN (total_width, allocation->width);
+  total_width = allocation->width;
   while (l != NULL)
     {
       WnckTask *task = WNCK_TASK (l->data);
@@ -968,6 +1131,12 @@ wnck_tasklist_update_lists (WnckTasklist *tasklist)
       app_task = WNCK_TASK (l->data);
 
       app_task->windows = g_list_sort (app_task->windows, wnck_task_compare);
+
+      /* so the number of windows in the task gets reset on the
+       * task label
+       */
+      wnck_task_update_visible_state (app_task);
+      
       
       l = l->next;
     }
@@ -1155,7 +1324,8 @@ wnck_tasklist_activate_task_window     (WnckTask *task)
 
       active_ws = wnck_screen_get_active_workspace (tasklist->priv->screen);
       window_ws = wnck_window_get_workspace (task->window);
-      if (active_ws != window_ws &&
+      if (window_ws &&
+          active_ws != window_ws &&
 	  !tasklist->priv->switch_workspace_on_unminimize)
 	wnck_workspace_activate (window_ws);
 	  
@@ -1170,6 +1340,12 @@ wnck_tasklist_activate_task_window     (WnckTask *task)
 	}
       else
 	{
+          WnckWorkspace *window_ws;
+          
+          window_ws = wnck_window_get_workspace (task->window);
+          if (window_ws)
+            wnck_workspace_activate (window_ws);
+
 	  wnck_window_activate (task->window);
 	}
     }
@@ -1186,7 +1362,8 @@ wnck_tasklist_activate_task_window     (WnckTask *task)
 
 
 static void
-wnck_task_popup_menu (WnckTask  *task)
+wnck_task_popup_menu (WnckTask *task,
+                      gboolean  action_submenu)
 {
   GtkWidget *menu;
   WnckTask *win_task;
@@ -1235,11 +1412,15 @@ wnck_task_popup_menu (WnckTask  *task)
 	}
       
       gtk_widget_show (menu_item);
-      
-      g_signal_connect_object (G_OBJECT (menu_item), "activate",
-                               G_CALLBACK (wnck_task_menu_activated),
-                               G_OBJECT (win_task),
-                               0);      
+
+       if (action_submenu)
+         gtk_menu_item_set_submenu (GTK_MENU_ITEM (menu_item),
+                                    wnck_create_window_action_menu (win_task->window));
+       else
+         g_signal_connect_object (G_OBJECT (menu_item), "activate",
+                                  G_CALLBACK (wnck_task_menu_activated),
+                                  G_OBJECT (win_task),
+                                  0);
       
       gtk_menu_shell_append (GTK_MENU_SHELL (menu), menu_item);
       
@@ -1271,7 +1452,7 @@ wnck_task_button_toggled (GtkButton *button,
 
   if (task->is_application)
     {
-      wnck_task_popup_menu (task);
+      wnck_task_popup_menu (task, FALSE);
     }
   else
     {
@@ -1295,7 +1476,7 @@ wnck_task_get_text (WnckTask *task)
     {
       return g_strdup_printf ("%s (%d)",
                               wnck_application_get_name (task->application),
-                              wnck_application_get_n_windows (task->application));
+                              g_list_length (task->windows));
     }
   else
     {
@@ -1504,21 +1685,37 @@ wnck_task_button_press_event (GtkWidget	      *widget,
 			      gpointer         data)
 {
   WnckTask *task = WNCK_TASK (data);
+
+  if (task->is_application)
+    {
+      wnck_task_popup_menu (task,
+                            event->button == 3);
+      return TRUE;
+    }
+  else if (event->button == 3)
+    {
+      if (task->action_menu)
+        gtk_widget_destroy (task->action_menu);
+
+      g_assert (task->action_menu == NULL);
+      
+      task->action_menu = wnck_create_window_action_menu (task->window);
+
+      g_object_add_weak_pointer (G_OBJECT (task->action_menu),
+                                 (void**) &task->action_menu);
+      
+      gtk_widget_show (task->action_menu);
+      gtk_menu_popup (GTK_MENU (task->action_menu),
+                      NULL, NULL,
+                      wnck_task_position_menu, task->button,
+                      event->button,
+                      gtk_get_current_event_time ());
+
+      /* FIXME we should really arrange to destroy the menu on popdown */
+      
+      return TRUE;
+    }
   
-  if (event->button != 1)
-    {
-      g_signal_stop_emission_by_name (widget,
-				      "button_press_event");
-    }
-  else
-    {
-      if (task->is_application)
-	{
-	  wnck_task_popup_menu (task);
-	  g_signal_stop_emission_by_name (widget,
-					  "button_press_event");
-	}
-    }
   return FALSE;
 }
 
